# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import os

from langchain_openai import ChatOpenAI

from swelldb.llm.abstract_llm import AbstractLLM
from swelldb.util.config_parser import ConfigParser
from swelldb.util.globals import Globals


class OpenAILLM(AbstractLLM):
    def __init__(self, model: str, api_key: str = None, temperature: int = 0) -> None:
        self.api_key: str = ""

        if api_key:
            self.api_key = api_key
        elif os.getenv(Globals.OPENAI_API_KEY):
            self.api_key = os.getenv(Globals.OPENAI_API_KEY)

        llm = ChatOpenAI(
            api_key=self.api_key,
            temperature=temperature,
            model=model,
        )
        super().__init__(llm=llm)



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from langchain.prompts import ChatPromptTemplate
from langchain_core.messages import BaseMessage
import os
from jinja2 import Environment, FileSystemLoader

from swelldb.table_plan.layout import Layout

column_format_examples = """
content: A list of movies
schema: {{'movie_name': 'the movie name', 'description': 'generate a description of the movie'}}
tuples: [['movie_1_name'], ['movie_2_name'], ['movie_3_name']]

"columns": {
  "movie_name": ["LOTR", "Harry Potter"],
  "description": [
    "The Lord of the Rings (LOTR) is an epic fantasy saga...",
    "Harry Potter is a fantasy film series..."
  ]
}
"""

row_format_examples = """
content: A list of movies
schema: {{'movie_name', 'the movie name', 'rating': 'the IMDB rating'}}

{
  "rows": [
    ["The Empire Strikes Back", 8.7],
    ["Blade Runner", 8.1],
    ["Back to the Future", 8.5]
  ]
}

content: A list of punk rock bands from the 80s
schema: {{'band_name', '', 'country': 'the country of origin'}}
{
  "rows": [
    ["The Ramones", "USA"],
    ["The Clash", "UK"],
    ["Dead Kennedys", "USA"]
  ]
}
"""

column_format_examples_with_data = """
content: None
schema: {superhero_name: the name, production_company: the comic production company}
tuples: {superman, spiderman, captain america}

"columns": {
  "superhero_name": ["Superman", "Spider-Man", "Captain America"],
  "production_company": ["DC Comics", "Marvel Comics", "Marvel Comics"]
}
"""

row_format_examples_with_data = """
content: None
schema: {superhero_name: the name, production_company: the comic production company}
tuples: {superman, spiderman, captain america}

"rows": [
    ["Superman", "DC Comics"],
    ["Spider-Man", "Marvel Comics"],
    ["Captain America", "Marvel Comics"]
  ]
"""


def generate_table_prompt(
    table_description: str, table_schema: dict[str, str]
) -> list[BaseMessage]:
    prompt: str = """\
    For the following table description, generate a table that contains rows with the following attributes:
    
    attributes: {attributes}
    
    description: {table_description}
    
    The format should be a key named 'rows', which maps to a list of rows.
    
    Produce the output in JSON format, such as it can by directly parsed in Python. Do not include ```json.
    """

    prompt_template: ChatPromptTemplate = ChatPromptTemplate.from_template(prompt)

    messages: list[BaseMessage] = prompt_template.format_messages(
        table_description=table_description, attributes=table_schema
    )

    return messages


def create_table_prompt(
    table_description: str, table_schema: dict[str, str], data: str, layout: Layout
):
    # Set up Jinja environment
    current_dir = os.path.dirname(os.path.abspath(__file__))
    prompts_dir = os.path.join(
        os.path.dirname(os.path.dirname(current_dir)),
        "swelldb",
        "table_plan",
        "prompts",
    )
    env = Environment(loader=FileSystemLoader(prompts_dir))

    # Load and render the template
    template = env.get_template("table_prompt.jinja")
    table_gen_prompt = template.render(
        table_description=table_description,
        table_schema=table_schema,
        data=data,
        layout=layout.get_name(),
    )

    return table_gen_prompt



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from langchain_ollama import ChatOllama

from swelldb.llm.abstract_llm import AbstractLLM


class OllamaLLM(AbstractLLM):
    def __init__(self, model):
        llm = ChatOllama(model=model, temperature=0)
        super().__init__(llm=llm)



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from typing import List

from pyarrow import Table

from swelldb.table_plan.table.physical.physical_table import PhysicalTable


class CustomTable(PhysicalTable):
    def __init__(self, table_name, data, chunk_size, layout, base_columns: List[str]):
        super().__init__(
            table_name,
            [],
            chunk_size=chunk_size,
            layout=layout,
            logical_table=None,
            child_table=None,
            operator_name=None,
            base_columns=base_columns,
        )

        self.data: Table = data

    def materialize(self, partitions: int = 1) -> Table:
        return self.data

    def __str__(self):
        return (
            f"CustomTable[schema={self.data.column_names}"
        )



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import os
from typing import Dict

from jinja2 import Template

from swelldb.table_plan.table.logical.logical_table import LogicalTable


@DeprecationWarning
def get_main_prompt(table: LogicalTable) -> str:
    prompt = f"""
Construct a table that contains the following content and schema:

content: {table.get_prompt()}
schema:  {table.get_schema()}
datasets: None

Your response should consist of the following JSON:

response structure: {plan_structure_prompt()}
    """

    return prompt


@DeprecationWarning
def plan_structure_prompt() -> str:
    structure = """
{    
    "dataset_columns": The columns that can be generated from any of the dataset. In dictionary format,
    of the form {dataset_name: {requested_column_name: original_column_name}}

    "missing_columns": The names of the columns of the requested schema that cannot be extracted from any of the dataset. 

    "llm_columns": The names of the columns that cannot be obtained from any of the given datasets or databases,
    but they are included in your knowledge.

    "search_engine_columns": The columns that you can generate only if you are provided with external data.

    "datasets": A list of the datasets you will use from the ones provided in the prompt, if any. Else return
    an empty list.

    "db_query": Generate a SQL query that extracts the data from the given database, if any. Otherwise return ''. 
    Use the appropriate aliases to comply with the requested schema. Your SQL query should also comply with the 
    schema of the given database.

    "linking_column": The most representative column for that entity of the given schema.
     Return this as a tuple of the form [column_name, dataset_name]. Use the column name according to the user-defined schema."
}
"""

    return structure


@DeprecationWarning
def get_last_line() -> str:
    return "\nReturn only the JSON response and no additional text. Do not start with ```json."


def get_local_tables_prompt(
    logical_table: LogicalTable, table_schema_dict: Dict[str, str]
):
    """
    Generate the prompt for the columns that can be retrieved from the local tables
    :param logical_table: The logical table
    :return: The prompt
    """

    # Get the directory of the current file
    current_dir = os.path.dirname(__file__)

    # Construct the relative path to the target file or directory
    prompt_file_path = os.path.join(
        current_dir, "prompts", "table_columns_prompt.jinja"
    )

    # Read the file and render the template
    with open(prompt_file_path, "r") as file:
        template = Template(file.read())
        prompt = template.render(
            content=logical_table.get_prompt(),
            schema=logical_table.get_schema().get_attribute_names(),
            tables=str(table_schema_dict),
        )

        return prompt


def get_llm_columns_prompt(logical_table: LogicalTable):
    """
    Generate the prompt for the LLM columns
    :param logical_table: The logical table
    :return: The prompt
    """

    # Get the directory of the current file
    current_dir = os.path.dirname(__file__)

    # Construct the relative path to the target file or directory
    prompt_file_path = os.path.join(
        current_dir, "prompts", "llm_table_columns_prompt.jinja"
    )

    # Read the file and render the template
    with open(prompt_file_path, "r") as file:
        template = Template(file.read())
        prompt = template.render(
            content=logical_table.get_prompt(),
            schema=logical_table.get_schema().get_attribute_names(),
        )

        return prompt



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import os

import pandas as pd
from pandas import DataFrame

from swelldb.datasource.datasource import Datasource


class LocalCSVDatasource(Datasource):
    def __init__(self, directory: str):
        self.directory = directory
        self.csv_files = [f for f in os.listdir(directory) if f.endswith(".csv")]

    def search(self, query: str) -> list[DataFrame]:
        results: dict[str, DataFrame] = dict()
        for csv_file in self.csv_files:
            results[csv_file.split(".")[0]] = pd.read_csv(
                os.path.join(self.directory, csv_file)
            )

        return results



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from typing import List
import os
import pyarrow as pa
from langchain_community.utilities import GoogleSerperAPIWrapper
from overrides import override
from jinja2 import Environment, FileSystemLoader

from swelldb.llm.abstract_llm import AbstractLLM
from swelldb.prompt.prompt_utils import create_table_prompt
from swelldb.table_plan.layout import Layout
from swelldb.table_plan.table.logical.logical_table import LogicalTable
from swelldb.table_plan.table.physical.physical_table import PhysicalTable
from swelldb.engine.execution_engine import ExecutionEngine

import logging

from swelldb.util.config_parser import ConfigParser
from swelldb.util.globals import Globals


class SearchEngineTable(PhysicalTable):
    def __init__(
        self,
        execution_engine: ExecutionEngine,
        logical_table: LogicalTable,
        child_table: PhysicalTable,
        base_columns: List[str],
        llm: AbstractLLM,
        layout: Layout = Layout.ROW(),
        serper_api_key: str = None,
    ):
        super().__init__(
            execution_engine=execution_engine,
            llm=llm,
            logical_table=logical_table,
            child_table=child_table,
            operator_name="search_engine_table",
        )

        self._execution_engine = execution_engine
        self._base_columns = base_columns
        self._llm = llm
        self._layout = layout
        self._serper_api_key = serper_api_key

        # Set up Jinja environment
        current_dir = os.path.dirname(os.path.abspath(__file__))
        prompts_dir = os.path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(current_dir))),
            "table_plan",
            "prompts",
        )
        self._env = Environment(loader=FileSystemLoader(prompts_dir))

        # Load config

        if serper_api_key:
            self._serper_api_key = serper_api_key
        elif os.getenv("SERPER_API_KEY"):
            self._serper_api_key = os.getenv("SERPER_API_KEY")
        else:
            self._serper_api_key = ConfigParser.get_config(
                Globals.GOOGLE_SERPER_API_KEY
            )

    def get_prompts(self, input_table: pa.Table) -> List[str]:
        # Search
        logging.info("Searching on the internet")

        data: List = list()

        if input_table:
            if self._base_columns:
                data = input_table.select(self._base_columns).to_pylist()
            else:
                data = input_table.to_pylist()

        # Load and render the Jinja template
        template = self._env.get_template("search_engine_prompt.jinja")
        search_query_prompt = template.render(
            prompt=self._logical_table.get_prompt(),
            sql_query=self._logical_table._sql_query,
            schema=self._logical_table.get_schema().get_attribute_names(),
            data=data,
        )

        search_queries: list[str] = self._llm.call(search_query_prompt).split("\n")

        logging.info(f"Search queries: {search_queries}")

        search: GoogleSerperAPIWrapper = GoogleSerperAPIWrapper(
            serper_api_key=self._serper_api_key
        )

        search_results: str = ""

        links = []

        for query in search_queries:
            logging.info(f"Issuing query: {query}")
            results: dict = search.results(query)
            parsed_results: str = str(results["organic"])
            search_results = search_results + "\n" + parsed_results

            for result in results["organic"]:
                link = result["link"]
                links.append(link)

        prompt: str = create_table_prompt(
            table_description=self._logical_table.get_prompt(),
            table_schema=self._logical_table.get_schema().get_attribute_names(),
            data=f"Original data: {data}\nSearch results: {search_results}",
            layout=self._layout,
        )

        logging.info(f"Table prompt: {prompt}")

        return [prompt]

    @override
    def __str__(self):
        return f"SearchEngineTable[schema={self._logical_table.get_schema().get_attribute_names()}"



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.


class Globals:
    """
    A class to hold global variables for the application.
    """

    # Global variables
    OPENAI_API_KEY: str = "openai_api_key"
    GOOGLE_SERPER_API_KEY: str = "serper_api_key"



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import os

from langchain_openai.chat_models.base import BaseChatOpenAI
from swelldb.llm.abstract_llm import AbstractLLM


class DeepseekOnlineLLM(AbstractLLM):
    def __init__(self, model="deepseek-chat"):
        llm = BaseChatOpenAI(
            model=model,
            openai_api_key=os.getenv("DEEPSEEK_API_KEY"),
            openai_api_base="https://api.deepseek.com",
            max_tokens=1024,
        )
        super().__init__(llm=llm)



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from typing import Dict, List

import pyarrow as pa
import sqlite3
import pandas as pd

from swelldb.datasource.database_datasource import DatabaseDatasource


class SQLiteDatasource(DatabaseDatasource):
    def __init__(self, path: str):
        self.path = path
        self.cursor = sqlite3.connect(path)

    def schema(self) -> Dict[str, List[str]]:
        res = self.cursor.execute("SELECT name FROM sqlite_master WHERE type='table'")
        tables = [row[0] for row in res.fetchall()]

        # Fetch column names for each table
        table_columns = {}
        for table in tables:
            res = self.cursor.execute(f"PRAGMA table_info({table})")
            columns = [
                row[1] for row in res.fetchall()
            ]  # Column names are in the second field
            table_columns[table] = columns

        return table_columns

    def sql(self, query: str) -> pa.Table:
        df = pd.read_sql_query(query, self.cursor)
        return pa.table(df)



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import configparser
import os

import yaml


class ConfigParser:
    def __init__(self, config_file):
        self.config_file = config_file
        self.config = {}

    @staticmethod
    def get_config(config_name: str) -> str:
        # Get current dir
        current_dir = os.path.dirname(__file__)
        # Get the path of the config file
        config_file = os.path.join(current_dir, "../../config", "config.yaml")

        with open(config_file, "r") as f:
            data = yaml.load(f, Loader=yaml.SafeLoader)
            return data["global"][config_name]



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import unittest
import pyarrow as pa

from swelldb.table_plan.swelldb_schema import SwellDBSchemaBuilder, SwellDBSchema
from swelldb.table_plan.table.logical.logical_table import LogicalTable


class TestSwellDBUsage(unittest.TestCase):
    def test_schema_1(self):
        builder: SwellDBSchemaBuilder = SwellDBSchemaBuilder()

        schema: SwellDBSchema = (
            builder.add_attribute("name", pa.string(), description="just a name")
            .add_attribute("age", pa.int32(), description=None)
            .add_attribute("text_field", pa.string(), description=None)
        ).build()

        self.assertEqual(schema.get_attribute("name").get_name(), "name")
        self.assertEqual(schema.get_attribute("age").get_name(), "age")
        self.assertEqual(schema.get_attribute("text_field").get_name(), "text_field")



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from typing import List, Dict
import os
import pyarrow as pa
from jinja2 import Environment, FileSystemLoader, Template
from overrides import overrides

from swelldb.engine.execution_engine import ExecutionEngine
from swelldb.llm.abstract_llm import AbstractLLM
from swelldb.prompt.prompt_utils import create_table_prompt
from swelldb.table_plan.layout import Layout
from swelldb.table_plan.swelldb_schema import SwellDBSchema
from swelldb.table_plan.table.logical.logical_table import LogicalTable
from swelldb.table_plan.table.physical.physical_table import PhysicalTable

import logging


class LLMTable(PhysicalTable):
    def __init__(
        self,
        execution_engine: ExecutionEngine,
        logical_table: LogicalTable,
        child_table: PhysicalTable,
        base_columns: List[str],
        llm: AbstractLLM,
        layout: Layout = Layout.ROW(),
    ):
        super().__init__(
            execution_engine=execution_engine,
            logical_table=logical_table,
            child_table=child_table,
            layout=layout,
            operator_name="llm_table",
            llm=llm,
            base_columns=base_columns,
        )

        # Set up Jinja environment
        current_dir = os.path.dirname(os.path.abspath(__file__))
        prompts_dir = os.path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(current_dir))),
            "table_plan",
            "prompts",
        )
        self._env = Environment(loader=FileSystemLoader(prompts_dir))

    def get_prompts(self, input_table: pa.Table) -> List[str]:
        logging.info("Generating LLM Table")

        data: List = list()

        if input_table:
            if self._base_columns:
                data = input_table.select(self._base_columns).to_pylist()
            else:
                data = input_table.to_pylist()

        schema: SwellDBSchema = self._logical_table.get_schema()

        prompt: str = create_table_prompt(
            table_description=self._logical_table.get_prompt(),
            table_schema=schema.get_attribute_names(),
            data=data,
            layout=self._layout,
        )

        return [prompt]

    @staticmethod
    def get_columns_prompt(logical_table: LogicalTable, tables: Dict[str, str]) -> str:
        # Get the directory of the current file
        current_dir = os.path.dirname(__file__)

        # Construct the relative path to the target file or directory
        prompt_file_path = os.path.join(
            current_dir, "../../prompts", f"llm_table_columns_prompt.jinja"
        )

        # Read the file and render the template
        with open(prompt_file_path, "r") as file:
            template = Template(file.read())
            prompt = template.render(
                content=logical_table.get_prompt(),
                schema=logical_table.get_schema().get_attribute_names(),
            )

            return prompt

    def __str__(self):
        return (
            f"LLMTable[schema={self._logical_table.get_schema().get_attribute_names()}"
        )



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from swelldb.table_plan.table.logical.logical_table import LogicalTable
from swelldb.engine.execution_engine import ExecutionEngine


import pyarrow as pa
import pyarrow.dataset as ds

from pandas import DataFrame
from datafusion import SessionContext, Catalog


class DataFusionEngine(ExecutionEngine):
    def __init__(self):
        self._sc: SessionContext = SessionContext()
        self._materialized_tables: dict[str, LogicalTable] = dict()

    def refresh(self) -> None:
        for table in self._materialized_tables.keys():
            self._sc.deregister_table(table)
        self._materialized_tables = dict()

    def get_tables(self):
        """
        Returns the table names and schemata
        :return:
        """
        tables = self._sc.tables()
        table_schemas = {}
        for table in tables:
            table_schemas[table] = (
                self._sc.table(table).schema().to_string().split("\n")
            )
        return table_schemas

    def register_csv(self, name: str, path: str):
        table_ds: ds.Dataset = ds.dataset(path, format="csv")

        if not self._sc.table_exist(name):
            self._sc.register_dataset(name, table_ds)

    def register_table(self, name: str, df: DataFrame):
        table_ds = ds.dataset(pa.Table.from_pandas(df))

        if not self._sc.table_exist(name):
            self._sc.register_dataset(name, table_ds)

    def deregister_table(self, name: str):
        self._sc.deregister_table(name)

    def sql(self, query: str) -> DataFrame:
        return self._sc.sql(query)



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from typing import List, Dict

from pyarrow.lib import DataType

import pyarrow as pa

from swelldb.table_plan.swelldb_attribute import SwellDBAttribute


class SwellDBSchema:
    def __init__(self, attributes: List[SwellDBAttribute]):
        self._attribute_list: List[SwellDBAttribute] = attributes
        self._attributes: Dict[str, SwellDBAttribute] = dict(
            [(attr.get_name(), attr) for attr in attributes]
        )

    def get_attributes(self) -> List[SwellDBAttribute]:
        return self._attribute_list

    def get_attribute(self, name: str) -> SwellDBAttribute:
        return self._attributes[name]

    def get_attribute_names(self) -> List[str]:
        return [d.get_name() for d in self._attribute_list]

    def to_arrow_schema(self) -> pa.Schema:
        return pa.schema(
            [
                (attr.get_name(), attr.get_data_type(), attr.get_description())
                for attr in self._attribute_list
            ]
        )

    @staticmethod
    def from_string(schema: str) -> "SwellDBSchema":
        attributes = []
        for attr in schema.split(","):
            attr_name, attr_type = attr.strip().split(" ")
            dt = pa.from_numpy_dtype(attr_type.strip())
            attributes.append(SwellDBAttribute(attr_name.strip(), dt))
        return SwellDBSchema(attributes)


class SwellDBSchemaBuilder:
    def __init__(self):
        self._attributes: Dict[str, SwellDBAttribute] = dict()

    def add_attribute(self, name: str, data_type: DataType, description: str):
        self._attributes[name] = SwellDBAttribute(
            name=name, dtype=data_type, description=description
        )

        return self

    def build(self):
        return SwellDBSchema(self._attributes.values())



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import asyncio
import json
from collections import defaultdict

import math
from typing import List, Dict

import pyarrow as pa
from pyarrow import Table
from pyarrow.lib import Schema, RecordBatch

from swelldb.engine.execution_engine import ExecutionEngine
from swelldb.llm.abstract_llm import AbstractLLM
from swelldb.table_plan.layout import Layout
from swelldb.table_plan.table.logical.logical_table import LogicalTable

import logging


class PhysicalTable:
    def __init__(
        self,
        execution_engine: ExecutionEngine,
        llm: AbstractLLM,
        logical_table: LogicalTable,
        child_table: "PhysicalTable",
        operator_name: str,
        layout: Layout = Layout.ROW(),
        base_columns: List[str] = None,
        chunk_size: int = 10,
    ):
        self._logical_table = logical_table
        self._chunk_size = chunk_size
        self._child_table = child_table
        self._layout: Layout = layout
        self._base_columns: str = base_columns
        self._operator_name: str = operator_name
        self._llm = llm
        self._execution_engine = execution_engine

    def get_prompts(self, input_table: pa.Table) -> List[str]:
        raise NotImplementedError()

    def get_operator_name(self) -> str:
        return self._operator_name

    # TODO: Resolve the issue with the search engine table
    @staticmethod
    def get_columns_prompt(logical_table: LogicalTable, tables: Dict[str, str]) -> str:
        return None

    def get_child_table(self) -> "PhysicalTable":
        return self._child_table

    def partition_table(self, data: pa.Table) -> List[pa.Table]:
        num_partitions: int = math.ceil(data.num_rows / self._chunk_size)
        partitions: List[pa.Table] = list()

        offset: int = 0
        for i in range(0, num_partitions):
            partition = data.slice(offset, self._chunk_size)
            partitions.append(partition)
            offset += self._chunk_size

        return partitions

    def materialize(self, partitions: int = 1) -> pa.Table:
        final_result: Table = None

        child_result: Table = (
            self._child_table.materialize(partitions) if self._child_table else None
        )

        prompts: List[str] = self.get_prompts(child_result)
        n_prompts: int = len(prompts)

        for idx, prompt in enumerate(prompts):
            logging.info("Processing prompt {}/{}".format(idx + 1, n_prompts))

            resp: str = self._llm.call(prompt)

            logging.info(f"Issuing LLM call with prompt: {prompt}")
            logging.info(f"Response: {resp}")

            # TODO: Create a method for that
            if self._layout == Layout.COLUMN():
                column_data: Dict = json.loads(resp)["columns"]
            elif self._layout == Layout.ROW():
                row_data: Dict = json.loads(resp)["rows"]
                column_data = defaultdict(list)
                for row in row_data:
                    for idx, attr in enumerate(
                        self._logical_table.get_schema().get_attributes()
                    ):
                        column_data[attr.get_name()].append(row[idx])

            # TODO Add sanity checks for the data schema

            # Cast the output to the correct schema
            output_tbl: pa.Table = pa.table(
                column_data, schema=self._logical_table.get_schema().to_arrow_schema()
            )

            result = output_tbl

            if child_result:
                result = result.join(
                    right_table=child_result,
                    keys=self._base_columns,
                    join_type="inner",
                )

            if not final_result:
                final_result = result
            else:
                final_result = pa.concat_tables(
                    [final_result, result.cast(final_result.schema)]
                )

        return final_result

    def explain(self, space="") -> None:
        print("{}{}".format(space, self.__str__()))
        if self._child_table:
            self._child_table.explain(space + "--")

    def __or__(self, other):
        other._child_table = self
        return other




# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.
import importlib
import os
from typing import List, Dict

from jinja2 import Template
from overrides import override, overrides
from pyarrow import Table

from swelldb.llm.abstract_llm import AbstractLLM
from swelldb.table_plan.layout import Layout
from swelldb.table_plan.table.logical.logical_table import LogicalTable
from swelldb.table_plan.table.physical.physical_table import PhysicalTable
from swelldb.engine.execution_engine import ExecutionEngine


class DatasetTable(PhysicalTable):
    def __init__(
        self,
        execution_engine: ExecutionEngine,
        logical_table: LogicalTable,
        child_table: PhysicalTable,
        base_columns: List[str],
        llm: AbstractLLM,
        layout: Layout = Layout.ROW(),
        query: str = None,
    ):
        super().__init__(
            logical_table=logical_table,
            child_table=child_table,
            layout=layout,
            operator_name="dataset_table",
            llm=llm,
            base_columns=base_columns,
            execution_engine=execution_engine,
        )

        self._execution_engine = execution_engine
        self._query = query

    @override
    def materialize(self, partitions=1) -> Table:
        tables = self._execution_engine.get_tables()

        if self._query:
            sql_query = self._query
        else:
            sql_query = (
                self._llm.call(
                    f"""
            You have access to the following table schemas:
            f{tables}
            
            Generate a SQL query to extract the data from the tables. Your target schema is the following:
            {self._logical_table.get_schema().get_attribute_names()}
            
            Use aliases if needed. Return only the SQL query as a python-compatibel text format.
            """
                )
                .replace("sql", "")
                .replace("```", "")
            )

        return self._execution_engine.sql(sql_query).to_arrow_table()

    @overrides
    def get_columns_prompt(logical_table: LogicalTable, tables: Dict[str, str]) -> str:
        # Get the directory of the current file
        current_dir = os.path.dirname(__file__)

        tables_str = ""

        for tbl in tables:
            tables_str += f"Table name:{tbl}: Schema: {tables[tbl]}\n"

        # Read the file and render the template
        with (
            importlib.resources.files("swelldb.table_plan.prompts")
            .joinpath("dataset_table_columns_prompt.jinja")
            .open("r", encoding="utf-8") as file
        ):
            template = Template(file.read())
            prompt = template.render(
                content=logical_table.get_prompt(),
                schema=logical_table.get_schema().get_attribute_names(),
                tables=tables_str,
            )

            return prompt

    @override
    def __str__(self):
        return f'DatasetTable[schema={self._logical_table.get_schema().get_attribute_names()}"]'



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import pyarrow as pa
from pyarrow.lib import DataType


class SwellDBAttribute:
    def __init__(self, name: str, dtype: DataType, description: str = None):
        self._name: str = name
        self._dtype: DataType = dtype
        self._description = description

    def get_name(self) -> str:
        return self._name

    def get_data_type(self) -> DataType:
        return self._dtype

    def get_description(self) -> str:
        return self._description



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from datafusion import DataFrame


class ExecutionEngine:
    def __init__(self):
        pass

    def sql(self, query: str) -> DataFrame:
        pass

    def get_tables(self):
        pass

    def register_csv(self, name: str, path: str):
        pass

    def register_table(self, name: str, df: DataFrame):
        pass

    def deregister_table(self, name: str):
        pass



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import json

from langchain_core.language_models import BaseChatModel


class AbstractLLM:
    def __init__(self, llm: BaseChatModel):
        self.llm: BaseChatModel = llm

        # Stats
        self.input_tokens = 0
        self.output_tokens = 0

    def call(self, prompt: str) -> str:
        r = self.llm.invoke(prompt)
        stats = r.usage_metadata

        self.input_tokens += stats["input_tokens"]
        self.output_tokens += stats["output_tokens"]

        r = r.content

        if "```json" in r:
            r = r.split("```json")[1].split("```")[0]
        else:
            r = r

        # For Deepseek responses
        if "</think>" in r:
            r = r.split("</think>")[1]

        return r

    def needs_search(self, content: str) -> bool:
        # Ask if further information is needed
        ask_prompt: str = f"""\
        I need to construct a table that contains information about the following content:

        content: {content}

        Your response should be up-to-date. Does your data suffice to do that?
        
        If no, respond 0. If yes, respond 1.
        """

        ans: str = self.llm(ask_prompt).content

        needs_search: bool = int(ans) == 0

        return needs_search



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from typing import Union, List, Dict
from enum import Enum

import pyarrow as pa

from swelldb.engine.datafusion_processor import DataFusionEngine
from swelldb.table_plan.planner import TableGenPlanner
from swelldb.table_plan.swelldb_schema import SwellDBSchema
from swelldb.llm.abstract_llm import AbstractLLM
from swelldb.table_plan.layout import Layout
from swelldb.table_plan.table.logical.logical_table import LogicalTable
from swelldb.table_plan.table.physical.custom_table import CustomTable
from swelldb.table_plan.table.physical.llm_table import LLMTable
from swelldb.table_plan.table.physical.physical_table import PhysicalTable
from swelldb.table_plan.table.physical.search_engine_table import SearchEngineTable
from swelldb.engine.execution_engine import ExecutionEngine
from swelldb.llm.openai_llm import OpenAILLM


class TableBuilder:
    def __init__(self, swelldb_ctx: "SwellDB"):
        self._table_name: str = None
        self._content: str = None
        self._schema: Union[SwellDBSchema, str] = None
        self._base_columns: List[str] = None
        self._table_gen_mode: Mode = Mode.LLM
        self._child_table = None
        self._data: pa.Table = None
        self._operators: List[type] = []
        self._chunk_size = 20

        self.swelldb_ctx = swelldb_ctx
        self.csv_files: List[(str, str)] = []
        self.parquet_files: List[(str, str)] = []

    def set_table_name(self, name: str) -> "TableBuilder":
        self._table_name = name
        return self

    def set_content(self, content: str) -> "TableBuilder":
        self._content = content
        return self

    def set_schema(self, schema: Union[SwellDBSchema, str]) -> "TableBuilder":
        self._schema = schema
        return self

    def set_base_columns(self, base_columns: List[str]) -> "TableBuilder":
        self._base_columns = base_columns
        return self

    def set_table_gen_mode(self, mode: "Mode") -> "TableBuilder":
        self._table_gen_mode = mode
        return self

    def set_operators(self, operators: List[type]):
        op_set = set()

        for operator in operators:
            if operator in op_set:
                raise ValueError(
                    f"Duplicate operator found: {operator}. Each operator should be included only once."
                )
            op_set.add(operator)

        self._operators = operators
        return self

    def set_data(self, data: pa.Table) -> "TableBuilder":
        if self._child_table:
            raise ValueError(
                "Cannot set data when child table is already set. Please use either data or child_table."
            )

        self._data = data
        return self

    def set_child_table(self, table: PhysicalTable) -> "TableBuilder":
        if self._data:
            raise ValueError(
                "Cannot set child table when data is already set. Please use either data or child_table."
            )

        self._child_table = table
        return self

    def set_chunk_size(self, chunk_size: int) -> "TableBuilder":
        self._chunk_size = chunk_size

        return self

    def add_csv_file(self, name: str, path: str) -> "TableBuilder":
        self.csv_files.append((name, path))
        return self

    def add_parquet_file(self, name: str, path: str) -> "TableBuilder":
        self.parquet_files.append((name, path))
        return self

    def build(self):
        """
        Build the table using the provided parameters.
        """
        if self._content is None:
            raise ValueError("Content must be set.")

        if self._schema is None:
            raise ValueError("Schema must be set.")

        for csv_file in self.csv_files:
            name, path = csv_file
            self.swelldb_ctx._execution_engine.register_csv(name=name, path=path)

        tables = self.swelldb_ctx._execution_engine.get_tables()

        return self.swelldb_ctx._create_table(
            name=self._table_name,
            content=self._content,
            schema=self._schema,
            chunk_size=self._chunk_size,
            operators=self._operators,
            base_columns=self._base_columns,
            mode=self._table_gen_mode,
            tables=tables,
            data=self._data,
            child_table=self._child_table,
        )


class Mode(Enum):
    PLANNER = "planner"
    OPERATORS = "operators"
    LLM = "llm"
    SEARCH = "search"
    DATASET = "dataset"


class SwellDB:
    def __init__(
        self,
        llm: AbstractLLM,
        execution_engine: ExecutionEngine = DataFusionEngine(),
        serper_api_key: str = None,
    ):
        self._execution_engine = execution_engine
        self._llm = llm
        self._serper_api_key = serper_api_key
        self._planner = TableGenPlanner(
            llm=llm, execution_engine=execution_engine, serper_api_key=serper_api_key
        )

    def table_builder(self) -> TableBuilder:
        """
        Returns a TableBuilder instance to build tables.
        """
        return TableBuilder(self)

    def _create_table(
        self,
        name: str,
        content: str,
        schema: Union[SwellDBSchema, str],
        chunk_size: int,
        base_columns: List[str] = None,
        mode: Mode = Mode.LLM,
        operators: List[type] = None,
        layout: Layout = Layout.ROW(),
        data: pa.Table = None,
        child_table: PhysicalTable = None,
        tables: Dict[str, str] = None,
    ) -> PhysicalTable:
        """
        Create a table using the given name, content, schema, layout and data.

        Args:
            name (str): The name of the table.
            content (str): The content of the table.
            schema (str): The schema of the table, as a string.
            base_columns (str): The base column to use for the table. Default is None.
            mode (Mode): The mode to use for table creation (LLM or SEARCH).
            layout (Layout): The layout of the table. Default is Layout.ROW().
            data (pa.Table): The data to be used for the table. Default is None.
            tables (Dict[str, str]): A dictionary of registered tables to be used for the table generation. Default is None.

        Returns:
            pa.Table: The created table.

        Examples:
            >>> from swelldb import SwellDB
            >>> from swelldb.llm.openai_llm import OpenAILLM

            >>> swell_ctx = SwellDB(OpenAILLM())
            >>> tbl = swell_ctx.create_table(
            ...     name="country",
            ...     content="a list of all US states",
            ...     schema="country_name, president, year",
            ...     layout=Layout.ROW(),
            ...     data=None
            ... )
        """

        if mode == Mode.PLANNER and not base_columns:
            raise ValueError(
                "Base columns (base_columns) must be specified in planner mode."
            )

        if mode == Mode.OPERATORS and not operators:
            raise ValueError("A list of operators should provider in operators mode.")

        if isinstance(schema, str):
            schema = SwellDBSchema.from_string(schema)

        # Create the logical table
        logical_table = LogicalTable(name=name, prompt=content, schema=schema)

        # If data is provided, create a child table that includes that data
        if data:
            child_table = CustomTable(
                "data",
                data=data,
                chunk_size=chunk_size,
                layout=layout,
                base_columns=base_columns,
            )

        # Experimental
        if mode == Mode.PLANNER:
            table: PhysicalTable = self._planner.create_plan(
                logical_table=logical_table, base_columns=base_columns, tables=tables
            )
        # Experimental
        elif mode == Mode.OPERATORS:
            table: PhysicalTable = self._planner.create_plan_from_operators(
                logical_table=logical_table,
                base_columns=base_columns,
                tables=tables,
                operators=operators,
            )
        elif mode == Mode.LLM:
            table: PhysicalTable = LLMTable(
                logical_table=logical_table,
                child_table=child_table,
                base_columns=base_columns,
                execution_engine=self._execution_engine,
                llm=self._llm,
                layout=layout,
            )
        elif mode == Mode.SEARCH:
            table: PhysicalTable = SearchEngineTable(
                logical_table=logical_table,
                child_table=child_table,
                base_columns=base_columns,
                execution_engine=self._execution_engine,
                llm=self._llm,
                layout=layout,
                serper_api_key=self._serper_api_key,
            )
        else:
            raise ValueError(f"Unknown mode: {mode}")

        return table



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.


class Datasource:
    def __init__(self):
        pass

    def search(self, query: str):
        pass



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from overrides import override
from pyarrow import Table

from swelldb.datasource.database_datasource import DatabaseDatasource
from swelldb.table_plan.table.logical.logical_table import LogicalTable
from swelldb.table_plan.table.physical.physical_table import PhysicalTable


class DatabaseTable(PhysicalTable):
    def __init__(
        self,
        logical_table: LogicalTable,
        query: str,
        database_datasource: DatabaseDatasource,
    ):
        super().__init__(logical_table=logical_table, children=[])

        self._query = query
        self._db = database_datasource

    @override
    def materialize(self) -> Table:
        return self._db.sql(self._query)

    @override
    def __str__(self):
        return f'DatabaseTable[schema={self._logical_table.get_schema()}, query="{self._query}"]'



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.


class Layout:
    def get_name(self) -> str:
        return self.__class__.__name__.upper()

    _ROW = None
    _COLUMN = None

    @classmethod
    def ROW(cls) -> "Layout":
        if cls._ROW is None:
            cls._ROW = Row()
        return cls._ROW

    @classmethod
    def COLUMN(cls) -> "Layout":
        if cls._COLUMN is None:
            cls._COLUMN = Column()
        return cls._COLUMN


class Column(Layout):
    pass


class Row(Layout):
    pass



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from typing import Dict

from swelldb.datasource.datasource import Datasource

import pyarrow as pa


class DatabaseDatasource(Datasource):
    def sql(self, query: str) -> pa.Table:
        raise NotImplementedError()

    def schema(self) -> Dict[str, str]:
        raise NotImplementedError()



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import pytest
from unittest.mock import Mock, patch
import pyarrow as pa
from swelldb import SwellDB
from swelldb.llm.openai_llm import OpenAILLM
from swelldb.swelldb import Mode
from swelldb.table_plan.layout import Layout
from swelldb.table_plan.swelldb_schema import SwellDBSchema
from swelldb.engine.datafusion_processor import DataFusionEngine


@pytest.fixture
def mock_llm():
    return Mock(spec=OpenAILLM)


@pytest.fixture
def mock_execution_engine():
    return Mock(spec=DataFusionEngine)


@pytest.fixture
def swelldb(mock_llm, mock_execution_engine):
    return SwellDB(
        llm=mock_llm, execution_engine=mock_execution_engine, serper_api_key="test_key"
    )


def test_swelldb_initialization(mock_llm, mock_execution_engine):
    db = SwellDB(
        llm=mock_llm, execution_engine=mock_execution_engine, serper_api_key="test_key"
    )
    assert db._llm == mock_llm
    assert db._execution_engine == mock_execution_engine
    assert db._serper_api_key == "test_key"


def test_create_table_llm_mode(swelldb, mock_llm):
    # Arrange
    name = "test_table"
    content = "test content"
    schema = SwellDBSchema.from_string("name str, age int")
    layout = Layout.ROW()

    # Act
    table = swelldb._create_table(
        name=name,
        content=content,
        schema=schema,
        mode=Mode.LLM,
        layout=layout,
        chunk_size=20,
    )

    # Assert
    assert table is not None
    mock_llm.call.assert_called()


def test_create_table_search_mode(swelldb, mock_llm):
    # Arrange
    name = "test_table"
    content = "test content"
    schema = SwellDBSchema.from_string("name str, age int")
    layout = Layout.ROW()

    # Act
    table = swelldb._create_table(
        name=name,
        content=content,
        schema=schema,
        mode=Mode.SEARCH,
        layout=layout,
        chunk_size=20,
    )

    # Assert
    assert table is not None
    mock_llm.call.assert_called()


def test_create_table_planner_mode(swelldb, mock_llm):
    # Arrange
    name = "test_table"
    content = "test content"
    schema = SwellDBSchema.from_string("name str, age int")
    layout = Layout.ROW()
    base_columns = ["name"]

    # Act
    table = swelldb._create_table(
        name=name,
        content=content,
        schema=schema,
        mode=Mode.PLANNER,
        layout=layout,
        base_columns=base_columns,
        chunk_size=20,
    )

    # Assert
    assert table is not None
    mock_llm.call.assert_called()


def test_create_table_invalid_mode(swelldb):
    # Arrange
    name = "test_table"
    content = "test content"
    schema = SwellDBSchema.from_string("name str, age int")
    layout = Layout.ROW()

    # Act & Assert
    with pytest.raises(ValueError, match="Unknown mode"):
        swelldb._create_table(
            name=name,
            content=content,
            schema=schema,
            mode="INVALID",
            layout=layout,
            chunk_size=20,
        )


def test_create_table_with_data(swelldb, mock_llm):
    # Arrange
    name = "test_table"
    content = "test content"
    schema = SwellDBSchema.from_string("name str, age int")
    layout = Layout.ROW()
    data = pa.table({"name": ["John", "Jane"], "age": [30, 25]})

    # Act
    table = swelldb._create_table(
        name=name,
        content=content,
        schema=schema,
        mode=Mode.LLM,
        layout=layout,
        data=data,
        chunk_size=20,
    )

    # Assert
    assert table is not None
    mock_llm.call.assert_called()


def test_create_table_planner_mode_missing_base_columns(swelldb):
    # Arrange
    name = "test_table"
    content = "test content"
    schema = SwellDBSchema.from_string("name str, age int")
    layout = Layout.ROW()

    # Act & Assert
    with pytest.raises(
        ValueError, match="Base columns must be specified in planner mode"
    ):
        swelldb._create_table(
            name=name,
            content=content,
            schema=schema,
            mode=Mode.PLANNER,
            layout=layout,
            chunk_size=20,
        )


@patch("swelldb.table_plan.planner.TableGenPlanner")
def test_planner_initialization(mock_planner, mock_llm, mock_execution_engine):
    # Arrange
    db = SwellDB(
        llm=mock_llm, execution_engine=mock_execution_engine, serper_api_key="test_key"
    )

    # Assert
    mock_planner.assert_called_once_with(
        llm=mock_llm, execution_engine=mock_execution_engine, serper_api_key="test_key"
    )



# Copyright (c) 2025 Victor Giannakouris
# 
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import os
import unittest

from swelldb import SwellDB
from swelldb.llm.openai_llm import OpenAILLM
from swelldb.table_plan.swelldb_schema import SwellDBSchema, SwellDBSchemaBuilder

import pyarrow as pa

import logging

from swelldb.swelldb import Mode, TableBuilder
from swelldb.table_plan.table.physical.dataset_table import DatasetTable
from swelldb.table_plan.table.physical.llm_table import LLMTable
from swelldb.table_plan.table.physical.physical_table import PhysicalTable
from swelldb.table_plan.table.physical.search_engine_table import SearchEngineTable

# Add info level logging
logging.basicConfig(level=logging.INFO)


class TestSwellDBPlanner(unittest.TestCase):
    def test_e2e_1(self):
        builder: SwellDBSchemaBuilder = SwellDBSchemaBuilder()

        schema: SwellDBSchema = (
            builder.add_attribute(
                "name", pa.string(), description="the state name"
            )
        ).build()

        swelldb: SwellDB = SwellDB(OpenAILLM(model="gpt-4o"))

        table_builder: TableBuilder = swelldb.table_builder()

        tbl: PhysicalTable = (
            table_builder.set_table_name("country")
            .set_content("a table of 5 US sates")
            .set_schema(schema)
            .set_base_columns(["name"])
            .set_table_gen_mode(Mode.OPERATORS)
            .set_operators([SearchEngineTable])
        ).build()

        print(tbl.materialize().to_pandas())

    def test_e2e_2(self):
        builder: SwellDBSchemaBuilder = SwellDBSchemaBuilder()

        schema: SwellDBSchema = (
            builder.add_attribute(
                "name", pa.string(), description="the state name"
            ).add_attribute(
                "governor_1995",
                pa.string(),
                description="the governor of that state in year 1995",
            )
        ).build()

        swelldb: SwellDB = SwellDB(OpenAILLM(model="gpt-4o"))

        table_builder: TableBuilder = swelldb.table_builder()

        tbl: PhysicalTable = (
            table_builder.set_table_name("country")
            .set_content("a table of 5 US sates")
            .set_schema(schema)
            .set_base_columns(["name"])
            .set_table_gen_mode(Mode.OPERATORS)
            .set_operators([SearchEngineTable])
        ).build()

        print(tbl.materialize().to_pandas())

    def test_e2e_3(self):
        swelldb: SwellDB = SwellDB(OpenAILLM(model="gpt-4o"))

        table_builder: TableBuilder = swelldb.table_builder()
        tbl: PhysicalTable = (
            table_builder.set_table_name("country")
            .set_content("a table of 5  US states")
            .set_schema("name str, population int")
            .set_base_columns(["name"])
            .set_table_gen_mode(Mode.OPERATORS)
            .set_operators([SearchEngineTable])
        ).build()

        print(tbl.materialize().to_pandas())

    def test_e2e_4(self):
        swelldb: SwellDB = SwellDB(OpenAILLM(temperature=0, model="gpt-4o"))

        # Create a dataframe with some states
        data: pa.Table = pa.Table.from_pydict(
            {"name": ["New York", "Illinois"], "state": ["NY", "IL"]}
        )

        table_builder: TableBuilder = swelldb.table_builder()
        tbl: PhysicalTable = (
            table_builder.set_table_name("country")
            .set_content("a table of US states")
            .set_schema("name str, state_population_1990 int, governor_1990 str")
            .set_base_columns(["name"])
            .set_data(data)
        ).build()

        tbl.explain()

        print(tbl.materialize().to_pandas())

    def test_e2e_builder(self):
        swelldb: SwellDB = SwellDB(OpenAILLM(temperature=0, model="gpt-4o"))
        table_builder: TableBuilder = swelldb.table_builder()

        file_dir = os.path.dirname(os.path.abspath(__file__))

        tbl: PhysicalTable = (
            table_builder.set_table_name("country")
            .set_content("a table of mutations")
            .set_schema("mut_name str, mutation_explanation str, web_link str")
            .set_table_gen_mode(Mode.OPERATORS)
            .set_operators([DatasetTable, LLMTable, SearchEngineTable])
            .set_base_columns(["mut_name"])
            .add_csv_file("mutations", os.path.join(file_dir, "test_files", "mutations.csv"))
            .add_csv_file("protein", os.path.join(file_dir, "test_files", "mutation_affected_protein.csv"))
        ).build()

        tbl.explain()

        print(tbl.materialize().to_pandas())

    def test_e2e_multi_key(self):
        swelldb: SwellDB = SwellDB(OpenAILLM(temperature=0, model="gpt-4o"))
        table_builder: TableBuilder = swelldb.table_builder()

        data: pa.Table = pa.Table.from_pydict(
            {
                "body_group": ["Upper", "Upper", "Lower"],
                "body_subgroup": ["Chest", "Back", "Legs"],
            }
        )

        tbl: PhysicalTable = (
            table_builder.set_table_name("exercises")
            .set_content("a table of exercises")
            .set_schema("body_group str, body_subgroup str, exercise_name str")
            .set_table_gen_mode(Mode.OPERATORS)
            .set_operators([LLMTable, SearchEngineTable])
            .set_base_columns(["body_group", "body_subgroup"])
            .set_data(data)
        ).build()

        tbl.explain()

        print(tbl.materialize().to_pandas())


    def test_bug(self):
        import pandas as pd
        import pyarrow as pa
        import sqlite3
        import os

        conn = sqlite3.connect("../examples/dsa.db")

        # Get all combinations
        combos = """
        SELECT 
          qt.id as question_type_id,
          qt.name as question_type_name,
          ql.id as question_level_id,
          ql.name as question_level_name
        FROM question_type as qt
        CROSS JOIN question_level as ql
        """

        combos = pd.read_sql(combos, conn)
        data = pa.Table.from_pandas(combos)

        swelldb: SwellDB = SwellDB(OpenAILLM(temperature=0, model="gpt-4o"))
        table_builder: TableBuilder = swelldb.table_builder()

        tbl = (
            swelldb.table_builder()
            .set_table_name("question")
            .set_content(
                "A table that contains DSA questions from Leetcode. Given the input data, create as many as possible.")
            .set_schema(
                "name str, link str, question_type_name str, question_type_id int, question_level_name str, question_level_id int")
            .set_base_columns(["question_type_name", "question_level_name"])
            .set_table_gen_mode(Mode.OPERATORS)
            .set_operators([LLMTable, SearchEngineTable])
            .set_data(data)
            .set_chunk_size(20)
        ).build()

        tbl.explain()

        # print(tbl.materialize().to_pandas())


# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import json

from typing import List, Set, Dict, Union

from swelldb.engine.execution_engine import ExecutionEngine
from swelldb.llm.abstract_llm import AbstractLLM
from swelldb.table_plan import planner_prompts
from swelldb.table_plan.swelldb_schema import SwellDBSchema

from swelldb.table_plan.table.logical.logical_table import LogicalTable
from swelldb.table_plan.table.physical.dataset_table import DatasetTable
from swelldb.table_plan.table.physical.llm_table import LLMTable
from swelldb.table_plan.table.physical.physical_table import PhysicalTable
from swelldb.table_plan.table.physical.search_engine_table import SearchEngineTable


class TableGenPlanner:
    def __init__(
        self, llm: AbstractLLM, execution_engine: ExecutionEngine, serper_api_key: str
    ):
        self._llm: AbstractLLM = llm
        self._execution_engine: ExecutionEngine = execution_engine
        self._serper_api_key: str = serper_api_key

    def create_plan_from_operators(
        self,
        logical_table: LogicalTable,
        base_columns: List[str],
        tables: Dict[str, str],
        operators: List[type],
    ):
        # The initial column set, defined by the user
        initial_schema: SwellDBSchema = logical_table.get_schema()

        # The remaining column set — Keeps track of the columns that are not yet generated
        remaining_column_set: Set[str] = set(initial_schema.get_attribute_names())

        root = None
        while operators and remaining_column_set:
            curr_logical_table_schema = SwellDBSchema(
                attributes=[
                    logical_table.get_schema().get_attribute(col)
                    for col in remaining_column_set
                ]
            )

            curr_logical_table = LogicalTable(
                name="tbl",
                prompt=logical_table.get_prompt(),
                schema=curr_logical_table_schema,
            )

            operator_cls = operators.pop(0)
            prompt: str = operator_cls.get_columns_prompt(curr_logical_table, tables)

            if not prompt:
                operator_columns = list(remaining_column_set)
            else:
                operator_columns = [
                    c.strip() for c in self._llm.call(prompt).split(",")
                ]

            for base_column in base_columns:
                if base_column not in operator_columns:
                    operator_columns.append(base_column)

            new_logical_table_schema = SwellDBSchema(
                attributes=[
                    logical_table.get_schema().get_attribute(col)
                    for col in operator_columns
                ]
            )

            new_logical_table = LogicalTable(
                name="tbl",
                prompt=logical_table.get_prompt(),
                schema=new_logical_table_schema,
            )

            # Replace the root table with the new operator — Add the previous root as its child.
            root = operator_cls(
                self._execution_engine, new_logical_table, root, base_columns, self._llm
            )

            remaining_column_set.difference_update(operator_columns)

        return root

    def create_plan(
        self,
        logical_table: LogicalTable,
        base_columns: List[str],
        tables: Dict[str, str] = dict(),
    ) -> PhysicalTable:

        # Root table operator
        root_table: PhysicalTable = None

        # The initial column set, defined by the user
        initial_schema: SwellDBSchema = logical_table.get_schema()

        # The remaining column set — Keeps track of the columns that are not yet generated
        remaining_column_set: Set[str] = set(initial_schema.get_attribute_names())

        # Get the columns that we can generate with the provided datasets
        local_ds_columns_response: str = self._llm.call(
            planner_prompts.get_local_tables_prompt(
                logical_table, table_schema_dict=tables
            )
        )

        # This response is a JSON string that contains the columns and the SQL query
        local_ds_columns_meta: Dict = json.loads(local_ds_columns_response)
        local_ds_columns = local_ds_columns_meta["columns"]
        local_ds_sql_query = local_ds_columns_meta["query"]

        # If any of the columns can be generated by the LLM
        if local_ds_columns_meta:
            local_ds_table_schema = SwellDBSchema(
                attributes=[
                    logical_table.get_schema().get_attribute(col)
                    for col in local_ds_columns
                ]
            )

            local_ds_logical_table = LogicalTable(
                name="search_table",
                prompt=logical_table.get_prompt(),
                schema=local_ds_table_schema,
            )

            # Set the root table
            root_table = DatasetTable(
                logical_table=local_ds_logical_table,
                llm=self._llm,
                base_columns=base_columns,
                child_table=None,
                query=local_ds_sql_query,
                execution_engine=self._execution_engine,
            )

            # Remove the columns that can be generated with the provided datasets from the remaining column set
            remaining_column_set = remaining_column_set.difference(local_ds_columns)

        # Check which columns can be generated by the LLM
        if remaining_column_set:
            tmp_logical_table = LogicalTable(
                name="llm_table",
                prompt=logical_table.get_prompt(),
                schema=SwellDBSchema(
                    attributes=[
                        logical_table.get_schema().get_attribute(col)
                        for col in remaining_column_set
                    ]
                ),
            )

            # Columns that the LLM can generate
            llm_columns: str = self._llm.call(
                planner_prompts.get_llm_columns_prompt(tmp_logical_table)
            )

            print(f"LLM columns: {llm_columns}")

            if llm_columns:
                llm_columns_list: Set[str] = set(
                    [x.strip() for x in llm_columns.split(",")]
                )

                remaining_column_set.difference_update(llm_columns_list)

                # add all base columns to llm_columns
                for base_column in base_columns:
                    if base_column not in llm_columns_list:
                        llm_columns_list.add(base_column)

                llm_table_schema: SwellDBSchema = SwellDBSchema(
                    attributes=[
                        logical_table.get_schema().get_attribute(col)
                        for col in llm_columns_list
                    ]
                )

                llm_logical_table = LogicalTable(
                    name="llm_table",
                    prompt=logical_table.get_prompt(),
                    schema=llm_table_schema,
                )

                root_table = LLMTable(
                    llm=self._llm,
                    execution_engine=self._execution_engine,
                    logical_table=llm_logical_table,
                    child_table=root_table,
                    base_column=base_column,
                )

        if remaining_column_set:
            remaining_column_set.add(base_column)

            search_table_schema = SwellDBSchema(
                attributes=[
                    logical_table.get_schema().get_attribute(col)
                    for col in remaining_column_set
                ]
            )

            search_logical_table = LogicalTable(
                name="search_table",
                prompt=logical_table.get_prompt(),
                schema=search_table_schema,
            )

            root_table = SearchEngineTable(
                llm=self._llm,
                logical_table=search_logical_table,
                child_table=root_table,
                base_column=base_column,
                execution_engine=self._execution_engine,
                serper_api_key=self._serper_api_key,
            )

        return root_table



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import pandas as pd

from swelldb.table_plan.table.logical.logical_table import LogicalTable


class Catalog:
    def __init__(self):
        self._tables: dict[str, LogicalTable] = dict()

    def register_llm_table(self, name: str, prompt: str, schema: dict):
        table: LogicalTable = LogicalTable(name=name, prompt=prompt, schema=schema)

        self._tables[name] = table

    def get_tables(self) -> list[LogicalTable]:
        return self._tables.values()

    def get_tables_df(self) -> pd.DataFrame:
        return pd.DataFrame.from_dict(self._tables.items())



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import unittest

from swelldb.table_plan.swelldb_attribute import SwellDBAttribute
from swelldb.table_plan.swelldb_schema import SwellDBSchema, SwellDBSchemaBuilder

import pyarrow as pa


class TestSwellDBUsage(unittest.TestCase):
    def test_schema_1(self):
        builder: SwellDBSchemaBuilder = SwellDBSchemaBuilder()

        schema: SwellDBSchema = (
            builder.add_attribute("name", pa.string(), description="just a name")
            .add_attribute("age", pa.int32(), description=None)
            .add_attribute("text_field", pa.string(), description=None)
        ).build()

        self.assertEquals(len(schema.get_attributes()), 3)

        self.assertEquals(schema.get_attribute("name").get_name(), "name")
        self.assertEquals(schema.get_attribute("age").get_name(), "age")
        self.assertEquals(schema.get_attribute("text_field").get_name(), "text_field")

        self.assertEquals(schema.get_attribute("name").get_data_type(), pa.string())
        self.assertEquals(schema.get_attribute("age").get_data_type(), pa.int32())
        self.assertEquals(
            schema.get_attribute("text_field").get_data_type(), pa.string()
        )



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

"""
SwellDB - A novel database system leveraging Large Language Models for natural language querying.
"""

from swelldb.swelldb import SwellDB
from swelldb.table_plan.planner import TableGenPlanner
from swelldb.llm.openai_llm import OpenAILLM
from swelldb.table_plan.swelldb_schema import SwellDBSchemaBuilder, SwellDBSchema
from swelldb.catalog import Catalog

__version__ = "0.1.0"
__all__ = [
    "SwellDB",
    "OpenAILLM",
    "SwellDBSchemaBuilder",
    "SwellDBSchema",
    "Catalog",
]



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.
import json

import pytest
from unittest.mock import Mock, patch
import pyarrow as pa

from swelldb.engine.datafusion_processor import DataFusionEngine
from swelldb.table_plan.planner import TableGenPlanner
from swelldb.table_plan.table.logical.logical_table import LogicalTable
from swelldb.table_plan.swelldb_schema import SwellDBSchema


@pytest.fixture
def mock_llm():
    return Mock()


@pytest.fixture
def mock_logical_table():
    schema = SwellDBSchema.from_string("name str, age int")
    return LogicalTable(name="test_table", prompt="test content", schema=schema)


@pytest.fixture
def planner(mock_llm):
    return TableGenPlanner(
        llm=mock_llm, execution_engine=DataFusionEngine(), serper_api_key=None
    )


def test_planner_initialization(mock_llm):
    planner = TableGenPlanner(
        llm=mock_llm, execution_engine=DataFusionEngine(), serper_api_key=None
    )
    assert planner._llm == mock_llm


def test_create_plan(planner, mock_logical_table, mock_llm):
    # Arrange
    base_columns = ["name"]
    mock_llm.call.return_value = json.dumps({"columns": [], "query": None})

    # Act
    plan = planner.create_plan(
        logical_table=mock_logical_table, base_columns=base_columns
    )

    # Assert
    assert plan is not None
    mock_llm.call.assert_called_once()


def test_create_plan_with_invalid_response(planner, mock_logical_table, mock_llm):
    # Arrange
    base_column = "name"
    mock_llm.call.return_value = "invalid_json"

    # Act & Assert
    with pytest.raises(ValueError):
        planner.create_plan(logical_table=mock_logical_table, base_column=base_column)


def test_create_plan_with_missing_base_column(planner, mock_logical_table):
    # Act & Assert
    with pytest.raises(ValueError, match="Base column must be specified"):
        planner.create_plan(logical_table=mock_logical_table, base_column=None)


@patch("swelldb.table_plan.planner.TableGenPlanner._validate_plan")
def test_create_plan_with_validation(
    mock_validate, planner, mock_logical_table, mock_llm
):
    # Arrange
    base_column = "name"
    mock_llm.call.return_value = '{"plan": "test_plan"}'
    mock_validate.return_value = True

    # Act
    plan = planner.create_plan(
        logical_table=mock_logical_table, base_column=base_column
    )

    # Assert
    assert plan is not None
    mock_validate.assert_called_once()


def test_create_plan_with_complex_schema(planner, mock_llm):
    # Arrange
    schema = SwellDBSchema.from_string(
        """
        name str,
        age int,
        address {
            street str,
            city str,
            zip str
        }
    """
    )
    logical_table = LogicalTable(
        name="test_table", prompt="test content", schema=schema
    )
    base_column = "name"
    mock_llm.call.return_value = '{"plan": "test_plan"}'

    # Act
    plan = planner.create_plan(logical_table=logical_table, base_column=base_column)

    # Assert
    assert plan is not None
    mock_llm.call.assert_called_once()


def test_create_plan_with_data(planner, mock_logical_table, mock_llm):
    # Arrange
    base_column = "name"
    data = pa.table({"name": ["John", "Jane"], "age": [30, 25]})
    mock_llm.call.return_value = '{"plan": "test_plan"}'

    # Act
    plan = planner.create_plan(
        logical_table=mock_logical_table, base_column=base_column, data=data
    )

    # Assert
    assert plan is not None
    mock_llm.call.assert_called_once()



# Copyright (c) 2025 Victor Giannakouris
#
# This file is part of SwellDB and is licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from typing import List, Set, Dict

from swelldb.table_plan.swelldb_attribute import SwellDBAttribute
from swelldb.table_plan.swelldb_schema import SwellDBSchema


class LogicalTable:
    def __init__(
        self, name: str, prompt: str, schema: SwellDBSchema, sql_query: str = None
    ):
        self._name: str = name
        self._prompt: str = prompt
        self._schema: SwellDBSchema = schema
        self._sql_query: str = sql_query

    def get_name(self) -> str:
        return self._name

    def get_prompt(self) -> str:
        return self._prompt

    def get_schema(self) -> SwellDBSchema:
        return self._schema

    def split(self, schemas: List[SwellDBSchema]) -> List["LogicalTable"]:
        """
        Splits the table into multiple ones according to the input schemata. For each schema, a sub-table will be
        crated
        :param schemas: The input schemata
        :return: The list of the sub-tables
        """
        leftover_columns: Set[SwellDBAttribute] = set(self._schema.get_attributes())
        split: List[LogicalTable] = list()

        idx: int = 0
        for schema in schemas:
            for attribute in schema.get_attributes():
                leftover_columns.remove(attribute)

            split.append(
                LogicalTable(
                    name=f"{self._name}_{idx}", prompt=self._prompt, schema=schema
                )
            )
            idx += 1

        if leftover_columns:
            split.append(
                LogicalTable(
                    name=f"{self._name}_{idx}",
                    prompt=self._prompt,
                    schema=SwellDBSchema(leftover_columns),
                )
            )

        return split

    def __str__(self) -> str:
        return f"{self._name} - {self._schema} - {self._prompt}"
